'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deactivate = exports.activate = void 0;
const vscode = require("vscode");
const vscode_1 = require("vscode");
const path = require("path");
const fs = require('fs');
const editing_1 = require("./features/editing");
const repository_1 = require("./data/repository");
const listener_1 = require("./data/listener");
const editingType = 'bpmn-io.editing';
const COMMANDS = {
    EDIT_CMD: 'extension.bpmn-io.edit'
};
var repository = new repository_1.Repository();
function createPanel(context, uri, provider) {
    const editorColumn = (vscode.window.activeTextEditor &&
        vscode.window.activeTextEditor.viewColumn) ||
        vscode.ViewColumn.One;
    const panel = vscode.window.createWebviewPanel(editingType, getPanelTitle(uri, provider), editorColumn, getWebviewOptions(context, uri));
    // set content
    panel.webview.html = provider.provideTextDocumentContent(uri, panel.webview);
    // init file
    repository.newBPMNFile(uri);
    listener_1.Listener(repository);
    // set panel icons
    const { extensionPath } = context;
    panel.iconPath = {
        light: getUri(extensionPath, 'resources', 'icon_light.svg'),
        dark: getUri(extensionPath, 'resources', 'icon_dark.svg')
    };
    // handling messages from the webview content
    panel.webview.onDidReceiveMessage(message => {
        switch (message.command) {
            case 'saveContent':
                saveFile(uri, message.content);
                return;
            case 'openEditor':
                repository.openEditor(uri, message.id);
                return;
        }
    }, undefined, context.subscriptions);
    return { panel, resource: uri, provider };
}
function saveFile(uri, content) {
    const { fsPath: docPath } = uri.with({ scheme: 'vscode-resource' });
    fs.writeFileSync(docPath, content, { encoding: 'utf8' });
}
function refresh(editor) {
    const { resource, panel, provider } = editor;
    panel.webview.html =
        provider.provideTextDocumentContent(resource, panel.webview);
}
function autoSaveIfConfigured(editorPanel, expectedStates) {
    const config = vscode.workspace.getConfiguration();
    const autoSaveConfiguration = config.get('files.autoSave');
    // do not save changes if autosave option does not satisfy
    if (autoSaveConfiguration === 'off' ||
        expectedStates.indexOf(autoSaveConfiguration) < 0) {
        return;
    }
    sendMessage('saveFile', editorPanel.panel.webview);
}
function activate(context) {
    const openedPanels = [];
    const editingProvider = new editing_1.EditingProvider(context);
    const _revealIfAlreadyOpened = (uri, provider) => {
        const opened = openedPanels.find(panel => {
            const { resource, provider: panelProvider } = panel;
            return resource.fsPath === uri.fsPath && panelProvider === provider;
        });
        if (!opened) {
            return false;
        }
        opened.panel.reveal(opened.panel.viewColumn);
        return true;
    };
    const _registerPanel = (editorPanel) => {
        // on editor closed
        editorPanel.panel.onDidDispose(() => {
            openedPanels.splice(openedPanels.indexOf(editorPanel), 1);
            /**
             * Note @pinussilvestrus
             *
             * We currently can't retrieve when a webview panel will be closed to save changes
             * before, cf. https://github.com/bpmn-io/vs-code-bpmn-io/issues/71#issuecomment-637543431
             */
            // autosaveIfConfigured(editorPanel, ['onFocusChange', 'onWindowChange', 'afterDelay']);
        });
        // on editor visibility changed
        editorPanel.panel.onDidChangeViewState(() => {
            refresh(editorPanel);
            autoSaveIfConfigured(editorPanel, ['onFocusChange', 'onWindowChange']);
        });
        openedPanels.push(editorPanel);
    };
    const _registerCommands = () => {
        const { EDIT_CMD } = COMMANDS;
        vscode.commands.registerCommand(EDIT_CMD, (uri) => {
            const documentUri = getDocumentUri(uri);
            if (documentUri && !_revealIfAlreadyOpened(documentUri, editingProvider)) {
                const panel = createPanel(context, documentUri, editingProvider);
                _registerPanel(panel);
                return panel;
            }
        });
    };
    const _serializePanel = (provider) => {
        const viewType = editingType;
        if (vscode.window.registerWebviewPanelSerializer) {
            vscode.window.registerWebviewPanelSerializer(viewType, {
                deserializeWebviewPanel(panel, state) {
                    return __awaiter(this, void 0, void 0, function* () {
                        if (!state || !state.resourcePath) {
                            return;
                        }
                        const resource = vscode_1.Uri.parse(state.resourcePath);
                        panel.title = panel.title || getPanelTitle(resource, provider);
                        panel.webview.options = getWebviewOptions(context, resource);
                        panel.webview.html =
                            provider.provideTextDocumentContent(resource, panel.webview);
                        _registerPanel({ panel, resource, provider });
                    });
                }
            });
        }
    };
    _registerCommands();
    _serializePanel(editingProvider);
}
exports.activate = activate;
function deactivate() { }
exports.deactivate = deactivate;
// helper ///////
function getPanelTitle(uri, provider) {
    const prefix = 'Edit';
    return `${prefix}: ${path.basename(uri.fsPath)}`;
}
function getWebviewOptions(context, uri) {
    return {
        enableScripts: true,
        retainContextWhenHidden: true,
        localResourceRoots: getLocalResourceRoots(context, uri)
    };
}
function getLocalResourceRoots(context, resource) {
    const baseRoots = [vscode.Uri.file(context.extensionPath)];
    const folder = vscode.workspace.getWorkspaceFolder(resource);
    if (folder) {
        return baseRoots.concat(folder.uri);
    }
    if (!resource.scheme || resource.scheme === 'file') {
        return baseRoots.concat(vscode.Uri.file(path.dirname(resource.fsPath)));
    }
    return baseRoots;
}
function getUri(...p) {
    return vscode.Uri.file(path.join(...p));
}
function sendMessage(message, webview) {
    webview.postMessage(message);
}
function getDocumentUri(uri) {
    const activeEditor = vscode.window.activeTextEditor;
    return uri || (activeEditor === null || activeEditor === void 0 ? void 0 : activeEditor.document.uri);
}
//# sourceMappingURL=extension.js.map